#pragma once

#define _USE_MATH_DEFINES
#include <cmath>
#include <math.h>
#include <cstdlib>
#include <cassert>

#include <stdexcept>
#include <string>
#include <sstream>
#include <algorithm>
#include <vector>
#include <deque>
#include <list>
#include <set>
#include <map>

#include <boost/foreach.hpp>
#include <boost/progress.hpp>
#include <boost/timer.hpp>

#include <BWAPI.h>
#include <BWTA.h>

#include "Options.h"
#include "UABAssert.h"


typedef unsigned char common_UnitCountType;


typedef std::vector<BWAPI::Unit> UnitVector;

BWAPI::AIModule * __NewAIModule();


enum tacticType { HydraliskPushTactic, MutaliskHarassTac, DefendTactic, ScoutTac, tactictypeEnd };// ZerglingHarassTac = 1 priority from weak to strong.
enum openingStrategy { TwelveHatchMuta, NinePoolling, TenHatchMuta, NinePoollingMuta, FivePoolling};


enum developMode { Develop, Release };
extern developMode	curMode;



struct double2
{
	double x, y;

	double2() {}
	double2(const double2& p) : x(p.x), y(p.y) {}
	double2(double x, double y) : x(x), y(y) {}
	double2(const BWAPI::Position & p) : x(p.x), y(p.y) {}

	operator BWAPI::Position()				const { return BWAPI::Position(static_cast<int>(x), static_cast<int>(y)); }

	double2 operator + (const double2 & v)	const { return double2(x + v.x, y + v.y); }
	double2 operator - (const double2 & v)	const { return double2(x - v.x, y - v.y); }
	double2 operator * (double s)			const { return double2(x*s, y*s); }
	double2 operator / (double s)			const { return double2(x / s, y / s); }

	double dot(const double2 & v)			const { return x*v.x + y*v.y; }
	double lenSq()							const { return x*x + y*y; }
	double len()							const { return sqrt(lenSq()); }
	double2 normal()						const { return *this / len(); }

	void normalise() { double s(len()); x /= s; y /= s; }
	void rotate(double angle)
	{
		angle = angle*M_PI / 180.0;
		*this = double2(x * cos(angle) - y * sin(angle), y * cos(angle) + x * sin(angle));
	}

	double2 rotateReturn(double angle)
	{
		angle = angle*M_PI / 180.0;
		return double2(x * cos(angle) - y * sin(angle), y * cos(angle) + x * sin(angle));
	}
};

int GetIntFromString(const std::string &s);

// Color codes for drawing text on the screen.
const char yellow = '\x03';
const char white = '\x04';
const char darkRed = '\x06';   // dim
const char green = '\x07';
const char red = '\x08';
const char purple = '\x10';   // dim
const char orange = '\x11';
const char gray = '\x1E';   // dim
const char cyan = '\x1F';
